import jax.numpy as jnp
import jax.random as jr
import pytest

from gpjax import Prior
from gpjax.kernels import RBF, to_spectral
from gpjax.likelihoods import Bernoulli, Gaussian, Poisson
from gpjax.mean_functions import Zero
from gpjax.parameters.base import _initialise_hyperparams, complete, initialise


def test_complete():
    posterior = Prior(kernel=RBF()) * Gaussian()
    partial_params = {"lengthscale": jnp.array(1.0)}
    full_params = complete(partial_params, posterior)
    assert list(full_params.keys()) == sorted(["lengthscale", "variance", "obs_noise"])


def test_initialise():
    posterior = Prior(kernel=RBF()) * Gaussian()
    params = initialise(posterior)
    assert list(params.keys()) == sorted(["lengthscale", "variance", "obs_noise"])


@pytest.mark.parametrize("n", [1, 10])
def test_non_conjugate_initialise(n):
    posterior = Prior(kernel=RBF()) * Bernoulli()
    params = initialise(posterior, n)
    assert list(params.keys()) == sorted(["lengthscale", "variance", "latent"])
    assert params["latent"].shape == (n, 1)


def test_hyperparametr_initialise():
    params = _initialise_hyperparams(RBF(), Zero())
    assert list(params.keys()) == sorted(["lengthscale", "variance"])


@pytest.mark.parametrize("lik", [Bernoulli, Poisson, Gaussian])
def test_dtype(lik):
    posterior = Prior(kernel=RBF()) * lik()
    for k, v in initialise(posterior, 10).items():
        assert v.dtype == jnp.float64


def test_spectral():
    key = jr.PRNGKey(123)
    kernel = to_spectral(RBF(), 10)
    posterior = Prior(kernel=kernel) * Gaussian()
    params = initialise(key, posterior)
    assert list(params.keys()) == sorted(["basis_fns", "obs_noise", "lengthscale", "variance"])
    assert params["basis_fns"].shape == (10, 1)



from typing import Callable

import jax.numpy as jnp
from jax.scipy.linalg import solve_triangular
from multipledispatch import dispatch

from ..gps import SpectralPosterior
from ..types import Dataset
from ..utils import I, concat_dictionaries


@dispatch(SpectralPosterior)
def marginal_ll(
    gp: SpectralPosterior,
    transform: Callable,
    negative: bool = False,
) -> Callable:
    def mll(params: dict, training: Dataset, static_params: dict = None):
        params = transform(params)
        x, y = training.X, training.y
        if static_params:
            params = concat_dictionaries(params, static_params)
        m = gp.prior.kernel.num_basis
        phi = gp.prior.kernel._build_phi(x, params)
        A = (params["variance"] / m) * jnp.matmul(jnp.transpose(phi), phi) + params[
            "obs_noise"
        ] * I(2 * m)

        RT = jnp.linalg.cholesky(A)
        R = jnp.transpose(RT)

        RtiPhit = solve_triangular(RT, jnp.transpose(phi))
        # Rtiphity=RtiPhit*y_tr;
        Rtiphity = jnp.matmul(RtiPhit, y)

        out = (
            0.5
            / params["obs_noise"]
            * (jnp.sum(jnp.square(y)) - params["variance"] / m * jnp.sum(jnp.square(Rtiphity)))
        )
        n = x.shape[0]

        out += (
            jnp.sum(jnp.log(jnp.diag(R)))
            + (n / 2.0 - m) * jnp.log(params["variance"])
            + n / 2 * jnp.log(2 * jnp.pi)
        )
        constant = jnp.array(-1.0) if negative else jnp.array(1.0)
        return constant * out.reshape()

    return mll



import jax.numpy as jnp
import pytest

from gpjax.kernels.utils import scale, stretch


@pytest.mark.parametrize("scale_dim", [1, 2, 5])
def test_scale(scale_dim):
    base_param = jnp.ones(shape=(10, scale_dim))
    scaler = jnp.ones(shape=(scale_dim,)) * jnp.array(2.0)
    scaled_param = scale(base_param, scaler)
    assert base_param.shape == scaled_param.shape
    assert (scaled_param == jnp.array(0.5)).all()
    assert scaled_param.dtype == jnp.float64


@pytest.mark.parametrize("matrix_shape", [(10, 10), (10, 1), (1, 10), (10, 3), (3, 10)])
@pytest.mark.parametrize("stretch_factor", [0.1, 0.5, 1.0, 10.0])
def test_stretch(matrix_shape, stretch_factor):
    A = jnp.ones(shape=matrix_shape)
    stretched_A = stretch(A, stretch_factor)
    assert (stretched_A == stretch_factor).all()
    assert stretched_A.shape == A.shape == matrix_shape
    assert stretched_A.dtype == jnp.float64



import jax.numpy as jnp
import pytest

from gpjax.kernels import RBF, cross_covariance, gram, initialise
from gpjax.utils import I


@pytest.mark.parametrize("dim", [1, 2, 5])
def test_gram(dim):
    x = jnp.linspace(-1.0, 1.0, num=10).reshape(-1, 1)
    if dim > 1:
        x = jnp.hstack([x] * dim)
    kern = RBF()
    params = initialise(kern)
    gram_matrix = gram(kern, x, params)
    assert gram_matrix.shape[0] == x.shape[0]
    assert gram_matrix.shape[0] == gram_matrix.shape[1]


@pytest.mark.parametrize("n1", [3, 10, 20])
@pytest.mark.parametrize("n2", [3, 10, 20])
def test_cross_covariance(n1, n2):
    x1 = jnp.linspace(-1.0, 1.0, num=n1).reshape(-1, 1)
    x2 = jnp.linspace(-1.0, 1.0, num=n2).reshape(-1, 1)
    params = initialise(RBF())
    kernel_matrix = cross_covariance(RBF(), x2, x1, params)
    assert kernel_matrix.shape == (n1, n2)


@pytest.mark.parametrize("dim", [1, 2, 5])
@pytest.mark.parametrize("ell, sigma", [(0.1, 0.1), (0.5, 0.1), (0.1, 0.5), (0.5, 0.5)])
def test_pos_def(dim, ell, sigma):
    n = 30
    x = jnp.linspace(0.0, 1.0, num=n).reshape(-1, 1)
    if dim > 1:
        x = jnp.hstack((x) * dim)
    kern = RBF()
    params = {"lengthscale": jnp.array([ell]), "variance": jnp.array(sigma)}

    gram_matrix = gram(kern, x, params)
    jitter_matrix = I(n) * 1e-6
    gram_matrix += jitter_matrix
    min_eig = jnp.linalg.eigvals(gram_matrix).min()
    assert min_eig > 0


@pytest.mark.parametrize("dim", [1, 2, 5, 10])
def test_initialisation(dim):
    params = initialise(RBF(ndims=dim))
    assert list(params.keys()) == ["lengthscale", "variance"]
    assert all(params["lengthscale"] == jnp.array([1.0] * dim))
    assert params["variance"] == jnp.array([1.0])


def test_dtype():
    params = initialise(RBF())
    for k, v in params.items():
        assert v.dtype == jnp.float64



import jax.numpy as jnp
import pytest
from tensorflow_probability.substrates.jax import distributions as tfd

from gpjax.gps import Prior
from gpjax.kernels import RBF
from gpjax.likelihoods import Bernoulli
from gpjax.parameters import log_density
from gpjax.parameters.priors import evaluate_prior, prior_checks


@pytest.mark.parametrize("x", [-1.0, 0.0, 1.0])
def test_lpd(x):
    val = jnp.array(x)
    dist = tfd.Normal(loc=0.0, scale=1.0)
    lpd = log_density(val, dist)
    assert lpd is not None


def test_prior_evaluation():
    """
    Test the regular setup that every parameter has a corresponding prior distribution attached to its unconstrained
    value.
    """
    params = {
        "lengthscale": jnp.array([1.0]),
        "variance": jnp.array([1.0]),
        "obs_noise": jnp.array([1.0]),
    }
    priors = {
        "lengthscale": tfd.Gamma(1.0, 1.0),
        "variance": tfd.Gamma(2.0, 2.0),
        "obs_noise": tfd.Gamma(3.0, 3.0),
    }
    lpd = evaluate_prior(params, priors)
    assert pytest.approx(lpd) == -2.0110168


def test_none_prior():
    """
    Test that multiple dispatch is working in the case of no priors.
    """
    params = {
        "lengthscale": jnp.array([1.0]),
        "variance": jnp.array([1.0]),
        "obs_noise": jnp.array([1.0]),
    }
    lpd = evaluate_prior(params, None)
    assert lpd == 0.0


def test_incomplete_priors():
    """
    Test the case where a user specifies priors for some, but not all, parameters.
    """
    params = {
        "lengthscale": jnp.array([1.0]),
        "variance": jnp.array([1.0]),
        "obs_noise": jnp.array([1.0]),
    }
    priors = {
        "lengthscale": tfd.Gamma(1.0, 1.0),
        "variance": tfd.Gamma(2.0, 2.0),
    }
    lpd = evaluate_prior(params, priors)
    assert pytest.approx(lpd) == -1.6137061


def test_checks():
    incomplete_priors = {"lengthscale": jnp.array([1.0])}
    posterior = Prior(kernel=RBF()) * Bernoulli()
    priors = prior_checks(posterior, incomplete_priors)
    assert "latent" in priors.keys()
    assert "variance" not in priors.keys()


def test_check_needless():
    complete_prior = {
        "lengthscale": tfd.Gamma(1.0, 1.0),
        "variance": tfd.Gamma(2.0, 2.0),
        "obs_noise": tfd.Gamma(3.0, 3.0),
        "latent": tfd.Normal(loc=0.0, scale=1.0),
    }
    posterior = Prior(kernel=RBF()) * Bernoulli()
    priors = prior_checks(posterior, complete_prior)
    assert priors == complete_prior



from typing import Tuple

import jax.numpy as jnp
from jax.scipy.linalg import cho_factor, cho_solve
from multipledispatch import dispatch

from .types import Array


def I(n: int) -> Array:
    """
    Compute an n x n identity matrix.
    :param n: The size of of the matrix.
    :return: An n x n identity matrix.
    """
    return jnp.eye(n)


def concat_dictionaries(a: dict, b: dict) -> dict:
    """
    Append one dictionary below another. If duplicate keys exist, then the key-value pair of the second supplied
    dictionary will be used.
    """
    return {**a, **b}


def merge_dictionaries(base_dict: dict, in_dict: dict) -> dict:
    """
    This will return a complete dictionary based on the keys of the first matrix. If the same key should exist in the
    second matrix, then the key-value pair from the first dictionary will be overwritten. The purpose of this is that
    the base_dict will be a complete dictionary of values such that an incomplete second dictionary can be used to
    update specific key-value pairs.

    :param base_dict: Complete dictionary of key-value pairs.
    :param in_dict: Subset of key-values pairs such that values from this dictionary will take precedent.
    :return: A merged single dictionary.
    """
    for k, v in base_dict.items():
        if k in in_dict.keys():
            base_dict[k] = in_dict[k]
    return base_dict


def sort_dictionary(base_dict: dict) -> dict:
    """
    Sort a dictionary based on the dictionary's key values.

    :param base_dict: The unsorted dictionary.
    :return: A dictionary sorted alphabetically on the dictionary's keys.
    """
    return dict(sorted(base_dict.items()))


@dispatch(jnp.DeviceArray)
def standardise(x: jnp.DeviceArray) -> Tuple[jnp.DeviceArray, jnp.DeviceArray, jnp.DeviceArray]:
    """
    Standardise a given matrix such that values are distributed according to a unit normal random variable. This is
    primarily designed for standardising a training dataset.

    :param x: A matrix of unstandardised values
    :return: A matrix of standardised values
    """
    xmean = jnp.mean(x, axis=0)
    xstd = jnp.std(x, axis=0)
    return (x - xmean) / xstd, xmean, xstd


@dispatch(jnp.DeviceArray, jnp.DeviceArray, jnp.DeviceArray)
def standardise(
    x: jnp.DeviceArray, xmean: jnp.DeviceArray, xstd: jnp.DeviceArray
) -> jnp.DeviceArray:
    """
    Standardise a given matrix with respect to a given mean and standard deviation. This is primarily designed for
    standardising a test set of data with respect to the training data.

    :param x: A matrix of unstandardised values
    :param xmean: A precomputed mean vector
    :param xstd: A precomputed standard deviation vector
    :return: A matrix of standardised values
    """
    return (x - xmean) / xstd


def unstandardise(
    x: jnp.DeviceArray, xmean: jnp.DeviceArray, xstd: jnp.DeviceArray
) -> jnp.DeviceArray:
    """
    Unstandardise a given matrix with respect to a previously computed mean and standard deviation. This is designed
    for remapping a matrix back onto its original scale.

    :param x: A standardised matrix.
    :param xmean: A mean vector.
    :param xstd: A standard deviation vector.
    :return: A matrix of unstandardised values.
    """
    return (x * xstd) + xmean



import jax.numpy as jnp
from multipledispatch import dispatch

from ..gps import (ConjugatePosterior, NonConjugatePosterior, Posterior,
                   SpectralPosterior)
from ..kernels import Kernel, initialise
from ..likelihoods import Likelihood, initialise
from ..mean_functions import MeanFunction, initialise
from ..utils import concat_dictionaries, merge_dictionaries, sort_dictionary


def _initialise_hyperparams(kernel: Kernel, meanf: MeanFunction) -> dict:
    return concat_dictionaries(initialise(kernel), initialise(meanf))


@dispatch(jnp.DeviceArray, (ConjugatePosterior, SpectralPosterior))
def initialise(key: jnp.DeviceArray, gp: ConjugatePosterior) -> dict:
    meanf = initialise(gp.prior.mean_function)
    kernel = concat_dictionaries(initialise(key, gp.prior.kernel), meanf)
    all_params = concat_dictionaries(kernel, initialise(gp.likelihood))
    return sort_dictionary(all_params)


@dispatch((ConjugatePosterior, SpectralPosterior))
def initialise(gp: ConjugatePosterior) -> dict:
    hyps = _initialise_hyperparams(gp.prior.kernel, gp.prior.mean_function)
    all_params = concat_dictionaries(hyps, initialise(gp.likelihood))
    return sort_dictionary(all_params)


@dispatch((ConjugatePosterior, SpectralPosterior), object)
def initialise(gp: ConjugatePosterior, n_data):
    return sort_dictionary(initialise(gp))


@dispatch(NonConjugatePosterior, int)
def initialise(gp: NonConjugatePosterior, n_data: int) -> dict:
    hyperparams = _initialise_hyperparams(gp.prior.kernel, gp.prior.mean_function)
    likelihood = concat_dictionaries(hyperparams, initialise(gp.likelihood))
    latent_process = {"latent": jnp.zeros(shape=(n_data, 1))}
    return sort_dictionary(concat_dictionaries(likelihood, latent_process))


def complete(params: dict, gp: Posterior, n_data: int = None) -> dict:
    full_param_set = initialise(gp, n_data)
    return sort_dictionary(merge_dictionaries(full_param_set, params))



import warnings

import jax.numpy as jnp
from chex import dataclass
from jax.interpreters.ad import JVPTracer
from jax.interpreters.partial_eval import DynamicJaxprTracer
from multipledispatch import dispatch
from tensorflow_probability.substrates.jax import distributions as tfd

from ..gps import NonConjugatePosterior
from ..types import Array, NoneType


@dispatch(Array, tfd.Distribution)
def log_density(param: jnp.DeviceArray, density: tfd.Distribution) -> Array:
    return density.log_prob(param)


@dispatch(dict, NoneType)
def evaluate_prior(params: dict, priors: dict) -> Array:
    return jnp.array(0.0)


@dispatch(dict, dict)
def evaluate_prior(params: dict, priors: dict) -> Array:
    lpd = jnp.array(0)
    for param, val in priors.items():
        lpd += jnp.sum(log_density(params[param], priors[param]))
    return lpd


@dispatch(NonConjugatePosterior, dict)
def prior_checks(gp: NonConjugatePosterior, priors: dict) -> dict:
    if "latent" in priors.keys():
        latent_prior = priors["latent"]
        if latent_prior.name != "Normal":
            warnings.warn(
                f"A {latent_prior.name} distribution prior has been placed on the latent function. It is strongly afvised that a unit-Gaussian prior is used."
            )
        return priors
    else:
        priors["latent"] = tfd.Normal(loc=0.0, scale=1.0)
        return priors



from typing import Optional, Union

import jax.numpy as jnp
from chex import dataclass
from jax.interpreters.ad import JVPTracer
from jax.interpreters.partial_eval import DynamicJaxprTracer
from jax.interpreters.pxla import ShardedDeviceArray

NoneType = type(None)
Array = (jnp.ndarray, ShardedDeviceArray, jnp.DeviceArray, JVPTracer, DynamicJaxprTracer)
Arrays = Union[jnp.ndarray, ShardedDeviceArray, jnp.DeviceArray, JVPTracer, DynamicJaxprTracer]


@dataclass(repr=False)
class Dataset:
    X: Arrays
    y: Arrays = None

    def __post_init__(self):
        assert (
            self.X.ndim == 2
        ), f"2-dimensional training inputs are required. Current dimension: {self.X.ndim}."
        if self.y != None:
            assert (
                self.y.ndim == 2
            ), f"2-dimensional training outputs are required. Current dimension: {self.y.ndim}."
            assert (
                self.X.shape[0] == self.y.shape[0]
            ), f"Number of inputs must equal the number of outputs. \nCurrent counts:\n- X: {self.X.shape[0]}\n- y: {self.y.shape[0]}"

    def __repr__(self) -> str:
        return f"- Number of datapoints: {self.X.shape[0]}\n- Dimension: {self.X.shape[1]}"

    @property
    def n(self) -> int:
        return self.X.shape[0]

    @property
    def in_dim(self) -> int:
        return self.X.shape[1]

    @property
    def out_dim(self) -> int:
        return self.y.shape[1]



import jax.numpy as jnp


def scale(x: jnp.DeviceArray, factor: jnp.DeviceArray) -> jnp.DeviceArray:
    return x / factor


def stretch(K: jnp.DeviceArray, factor: jnp.DeviceArray) -> jnp.DeviceArray:
    return factor * K



from jax.config import config

config.update("jax_enable_x64", True)
from .gps import Prior
from .predict import mean, variance
from .sampling import random_variable, sample
from .types import Dataset

__version__ = "0.3.4"



import jax.numpy as jnp
import jax.random as jr
import pytest
from tensorflow_probability.substrates.jax import distributions as tfd

from gpjax.kernels import RBF
from gpjax.kernels.spectral import (SpectralRBF, initialise,
                                    sample_frequencies, spectral_density,
                                    to_spectral)


@pytest.mark.parametrize("n_basis", [1, 2, 10])
def test_initialise(n_basis):
    key = jr.PRNGKey(123)
    kernel = SpectralRBF(num_basis=n_basis)
    params = initialise(key, kernel)
    assert list(params.keys()) == ["basis_fns", "lengthscale", "variance"]
    for v in params.values():
        assert v.dtype == jnp.float64


@pytest.mark.parametrize("n_basis", [1, 2, 10])
def test_to_spectral(n_basis):
    base_kern = RBF()
    spectral = to_spectral(base_kern, n_basis)
    assert isinstance(spectral, SpectralRBF)
    assert spectral.num_basis == n_basis
    assert spectral.stationary


def test_spectral_density():
    kernel = SpectralRBF(num_basis=10)
    sdensity = spectral_density(kernel)
    assert isinstance(sdensity, tfd.Normal)


@pytest.mark.parametrize("n_freqs", [1, 2, 5])
def test_sample_frequencies(n_freqs):
    key = jr.PRNGKey(123)
    kernel = SpectralRBF(num_basis=n_freqs)
    sdensity = spectral_density(kernel)
    omega = sample_frequencies(key, kernel, n_freqs, 1)
    omegad = sample_frequencies(key, sdensity, n_freqs, 1)
    assert (omegad == omega).all()
    assert omegad.dtype == jnp.float64
    assert omega.dtype == jnp.float64



from typing import Callable

import jax.numpy as jnp
from multipledispatch import dispatch
from tensorflow_probability.substrates.jax import distributions as tfd

from ..gps import ConjugatePosterior, NonConjugatePosterior
from ..kernels import gram
from ..likelihoods import link_function
from ..parameters.priors import evaluate_prior, prior_checks
from ..types import Array, Dataset
from ..utils import I, concat_dictionaries


@dispatch(ConjugatePosterior)
def marginal_ll(
    gp: ConjugatePosterior,
    transform: Callable,
    negative: bool = False,
) -> Callable:
    r"""
    Compute :math:`\log p(y | x, \theta) for a conjugate, or exact, Gaussian process.
    Args:
        x: A set of N X M inputs
        y: A set of N X 1 outputs
    Returns: A multivariate normal distribution
    """

    def mll(params: dict, training: Dataset, priors: dict = None, static_params: dict = None):
        x, y = training.X, training.y
        params = transform(params)
        if static_params:
            params = concat_dictionaries(params, transform(static_params))
        mu = gp.prior.mean_function(x)
        gram_matrix = gram(gp.prior.kernel, x, params)
        gram_matrix += params["obs_noise"] * I(x.shape[0])
        L = jnp.linalg.cholesky(gram_matrix)
        random_variable = tfd.MultivariateNormalTriL(mu, L)

        log_prior_density = evaluate_prior(params, priors)
        constant = jnp.array(-1.0) if negative else jnp.array(1.0)
        return constant * (random_variable.log_prob(y.squeeze()).mean() + log_prior_density)

    return mll


@dispatch(NonConjugatePosterior)
def marginal_ll(
    gp: NonConjugatePosterior,
    transform: Callable,
    negative: bool = False,
    jitter: float = 1e-6,
) -> Callable:
    def mll(
        params: dict,
        training: Dataset,
        priors: dict = {"latent": tfd.Normal(loc=0.0, scale=1.0)},
        static_params: dict = None,
    ):
        x, y = training.X, training.y
        n = training.n
        params = transform(params)
        if static_params:
            params = concat_dictionaries(params, transform(static_params))
        link = link_function(gp.likelihood)
        gram_matrix = gram(gp.prior.kernel, x, params)
        gram_matrix += I(n) * jitter
        L = jnp.linalg.cholesky(gram_matrix)
        F = jnp.matmul(L, params["latent"])
        rv = link(F)
        ll = jnp.sum(rv.log_prob(y))

        priors = prior_checks(gp, priors)
        log_prior_density = evaluate_prior(params, priors)
        constant = jnp.array(-1.0) if negative else jnp.array(1.0)
        return constant * (ll + log_prior_density)

    return mll



from typing import Callable

import jax.numpy as jnp
import pytest
from numpy.testing import assert_array_equal

from gpjax import Prior
from gpjax.config import get_defaults
from gpjax.kernels import RBF
from gpjax.likelihoods import Bernoulli, Gaussian, Poisson
from gpjax.parameters import (build_all_transforms, build_constrain,
                              build_unconstrain, initialise)


@pytest.mark.parametrize("transformation", [build_constrain, build_unconstrain])
@pytest.mark.parametrize("likelihood", [Gaussian, Poisson, Bernoulli])
def test_output(transformation, likelihood):
    posterior = Prior(kernel=RBF()) * likelihood()
    params = initialise(posterior, 10)
    config = get_defaults()
    transform_map = transformation(params.keys(), config)
    assert isinstance(transform_map, Callable)


@pytest.mark.parametrize("likelihood", [Gaussian, Poisson, Bernoulli])
def test_constrain(likelihood):
    posterior = Prior(kernel=RBF()) * likelihood()
    params = initialise(posterior, 10)
    config = get_defaults()
    transform_map = build_constrain(params.keys(), config)
    transformed_params = transform_map(params)
    assert transformed_params.keys() == params.keys()
    for u, v in zip(transformed_params.values(), params.values()):
        assert u.dtype == v.dtype


@pytest.mark.parametrize("likelihood", [Gaussian, Poisson, Bernoulli])
def test_unconstrain(likelihood):
    posterior = Prior(kernel=RBF()) * likelihood()
    params = initialise(posterior, 10)
    config = get_defaults()
    constrain_map = build_constrain(params.keys(), config)
    unconstrain_map = build_unconstrain(params.keys(), config)
    transformed_params = unconstrain_map(constrain_map(params))
    assert transformed_params.keys() == params.keys()
    for u, v in zip(transformed_params.values(), params.values()):
        assert_array_equal(u, v)
        assert u.dtype == v.dtype


@pytest.mark.parametrize("likelihood", [Gaussian, Poisson, Bernoulli])
def test_build_all_transforms(likelihood):
    posterior = Prior(kernel=RBF()) * likelihood()
    params = initialise(posterior, 10)
    config = get_defaults()
    t1, t2 = build_all_transforms(params.keys(), config)
    constrainer = build_constrain(params.keys(), config)
    constrained = t1(params)
    constrained2 = constrainer(params)
    assert constrained2.keys() == constrained2.keys()
    for u, v in zip(constrained.values(), constrained2.values()):
        assert_array_equal(u, v)
        assert u.dtype == v.dtype
    unconstrained = t2(params)
    unconstrainer = build_unconstrain(params.keys(), config)
    unconstrained2 = unconstrainer(params)
    for u, v in zip(unconstrained.values(), unconstrained2.values()):
        assert_array_equal(u, v)
        assert u.dtype == v.dtype



from typing import Callable

import jax.numpy as jnp
import jax.random as jr
import pytest
from tensorflow_probability.substrates.jax import distributions as tfd

from gpjax import Dataset, Prior
from gpjax.config import get_defaults
from gpjax.kernels import RBF
from gpjax.likelihoods import Bernoulli, Gaussian
from gpjax.objectives import marginal_ll
from gpjax.parameters import (build_all_transforms, build_unconstrain,
                              initialise)


def test_conjugate():
    posterior = Prior(kernel=RBF()) * Gaussian()

    x = jnp.linspace(-1.0, 1.0, 20).reshape(-1, 1)
    y = jnp.sin(x)
    D = Dataset(X=x, y=y)
    params = initialise(posterior)
    config = get_defaults()
    unconstrainer, constrainer = build_all_transforms(params.keys(), config)
    params = unconstrainer(params)
    mll = marginal_ll(posterior, transform=constrainer)
    assert isinstance(mll, Callable)
    neg_mll = marginal_ll(posterior, transform=constrainer, negative=True)
    assert neg_mll(params, D) == jnp.array(-1.0) * mll(params, D)


def test_non_conjugate():
    posterior = Prior(kernel=RBF()) * Bernoulli()
    n = 20
    x = jnp.linspace(-1.0, 1.0, n).reshape(-1, 1)
    y = jnp.sin(x)
    D = Dataset(X=x, y=y)
    params = initialise(posterior, 20)
    config = get_defaults()
    unconstrainer, constrainer = build_all_transforms(params.keys(), config)
    params = unconstrainer(params)
    mll = marginal_ll(posterior, transform=constrainer)
    assert isinstance(mll, Callable)
    neg_mll = marginal_ll(posterior, transform=constrainer, negative=True)
    assert neg_mll(params, D) == jnp.array(-1.0) * mll(params, D)


def test_prior_mll():
    """
    Test that the MLL evaluation works with priors attached to the parameter values.
    """
    key = jr.PRNGKey(123)
    x = jnp.sort(jr.uniform(key, minval=-5.0, maxval=5.0, shape=(100, 1)), axis=0)
    f = lambda x: jnp.sin(jnp.pi * x) / (jnp.pi * x)
    y = f(x) + jr.normal(key, shape=x.shape) * 0.1
    D = Dataset(X=x, y=y)
    posterior = Prior(kernel=RBF()) * Gaussian()

    params = initialise(posterior)
    config = get_defaults()
    constrainer, unconstrainer = build_all_transforms(params.keys(), config)
    params = unconstrainer(params)
    print(params)

    mll = marginal_ll(posterior, transform=constrainer)

    priors = {
        "lengthscale": tfd.Gamma(1.0, 1.0),
        "variance": tfd.Gamma(2.0, 2.0),
        "obs_noise": tfd.Gamma(2.0, 2.0),
    }
    mll_eval = mll(params, D)
    mll_eval_priors = mll(params, D, priors)

    assert pytest.approx(mll_eval) == jnp.array(-103.28180663)
    assert pytest.approx(mll_eval_priors) == jnp.array(-105.509218857)



from collections.abc import KeysView
from typing import Callable, Tuple

import jax.numpy as jnp
from ml_collections import ConfigDict


def build_unconstrain(keys: KeysView, config: ConfigDict) -> Callable:
    """
    Build the transformation map the will transform a set of parameters such that each parameter is now defined on the
    entire real line.

    :param keys: A set of dictionary keys that represent parameters name-key.
    :param config: A configuration dictionary that informs which transformation should be applied to which parameter.
    :return: A callable that will apply the desired transformation(s).
    """
    transforms = {k: config.transformations[config.transformations[k]]() for k in keys}

    def unconstrain(params: dict) -> dict:
        return {k: jnp.array(transforms[k].inverse(v)) for k, v in params.items()}

    return unconstrain


def build_constrain(keys: KeysView, config: ConfigDict) -> Callable:
    """
    Build the transformation map the will transform a set of parameters such that each unconstrained parameter
    representation is now defined on the parameter's original, possibly constrained, space.

    :param keys: A set of dictionary keys that represent parameters name-key.
    :param config: A configuration dictionary that informs which transformation should be applied to which parameter.
    :return: A callable that will apply the desired transformation(s).
    """
    transforms = {k: config.transformations[config.transformations[k]]() for k in keys}

    def constrain(params: dict) -> dict:
        return {k: jnp.array(transforms[k].forward(v)) for k, v in params.items()}

    return constrain


def build_all_transforms(keys: KeysView, config: ConfigDict) -> Tuple[Callable, Callable]:
    """
    Build both the constraining and unconstraining function mappings.

    :param keys: A set of dictionary keys that represent parameters name-key.
    :param config: A configuration dictionary that informs which transformation should be applied to which parameter.
    :return: A tuple of callables that will apply the desired transformation(s) to both the original and the unconstrained parameter values, in this order.
    """
    unconstrain = build_unconstrain(keys, config)
    constrain = build_constrain(keys, config)
    return constrain, unconstrain



from typing import Optional

import jax.numpy as jnp
from chex import dataclass
from multipledispatch import dispatch
from tensorflow_probability.substrates.jax import distributions as tfd

from gpjax.kernels import RBF

from ..types import Array
from ..utils import I, sort_dictionary
from .base import Kernel
from .utils import scale, stretch


@dataclass(repr=False)
class SpectralKernel:
    num_basis: int


@dataclass(repr=False)
class SpectralRBF(Kernel, SpectralKernel):
    name: Optional[str] = "Spectral RBF"
    stationary: str = True

    def __repr__(self):
        return f"{self.name}:\n\t Number of basis functions: {self.num_basis}\n\t Stationary: {self.stationary} \n\t ARD structure: {self.ard}"

    def __call__(self, x: jnp.DeviceArray, y: jnp.DeviceArray, params: dict) -> Array:
        phi = self._build_phi(x, params)
        return jnp.matmul(phi, jnp.transpose(phi)) / self.num_basis

    def _build_phi(self, x: jnp.DeviceArray, params):
        scaled_freqs = scale(params["basis_fns"], params["lengthscale"])
        phi = jnp.matmul(x, jnp.transpose(scaled_freqs))
        return jnp.hstack([jnp.cos(phi), jnp.sin(phi)])


@dispatch(RBF, int)
def to_spectral(kernel: RBF, num_basis: int):
    return SpectralRBF(num_basis=num_basis)


@dispatch(SpectralRBF)
def spectral_density(kernel: SpectralRBF) -> tfd.Distribution:
    return tfd.Normal(loc=jnp.array(0.0), scale=jnp.array(1.0))


@dispatch(jnp.DeviceArray, tfd.Distribution, int, int)
def sample_frequencies(
    key, density: tfd.Distribution, n_frequencies: int, input_dimension: int
) -> jnp.DeviceArray:
    return density.sample(sample_shape=(n_frequencies, input_dimension), seed=key)


@dispatch(jnp.DeviceArray, SpectralKernel, int, int)
def sample_frequencies(
    key, kernel: SpectralKernel, n_frequencies: int, input_dimension: int
) -> jnp.DeviceArray:
    density = spectral_density(kernel)
    return density.sample(sample_shape=(n_frequencies, input_dimension), seed=key)


@dispatch(jnp.DeviceArray, SpectralRBF)
def initialise(key: jnp.DeviceArray, kernel: SpectralRBF):
    basis_init = sample_frequencies(key, kernel, kernel.num_basis, kernel.ndims)
    return {
        "basis_fns": basis_init,
        "lengthscale": jnp.array([1.0] * kernel.ndims),
        "variance": jnp.array([1.0]),
    }



from typing import Callable

import jax.numpy as jnp
from jax.scipy.linalg import solve_triangular
from multipledispatch import dispatch
from tensorflow_probability.substrates.jax import distributions as tfd

from .gps import (ConjugatePosterior, NonConjugatePosterior, Prior,
                  SpectralPosterior)
from .kernels import cross_covariance, gram
from .likelihoods import predictive_moments
from .predict import mean, variance
from .types import Array, Dataset
from .utils import I, concat_dictionaries


@dispatch(Prior, dict, Dataset)
def random_variable(
    gp: Prior, params: dict, sample_points: Dataset, jitter_amount: float = 1e-6
) -> tfd.Distribution:
    X = sample_points.X
    N = sample_points.n
    mu = gp.mean_function(X)
    gram_matrix = gram(gp.kernel, X, params)
    jitter_matrix = I(N) * jitter_amount
    covariance = gram_matrix + jitter_matrix
    return tfd.MultivariateNormalFullCovariance(mu.squeeze(), covariance)


@dispatch(ConjugatePosterior, dict, Dataset)
def random_variable(
    gp: ConjugatePosterior,
    params: dict,
    training: Dataset,
    jitter_amount: float = 1e-6,
) -> Callable:
    # TODO: Return kernel matrices here to avoid replicated computation.
    meanf = mean(gp, params, training)
    covf = variance(gp, params, training)

    def build_rv(test_points: Array):
        n = test_points.shape[0]
        mu = meanf(test_points)
        cov = covf(test_points)
        return tfd.MultivariateNormalFullCovariance(mu.squeeze(), cov + I(n) * jitter_amount)

    return build_rv


@dispatch(NonConjugatePosterior, dict, Dataset)
def random_variable(
    gp: NonConjugatePosterior,
    params: dict,
    training: Dataset,
) -> Callable:
    nu = params["latent"]
    N = training.n
    X, y = training.X, training.y
    Kff = gram(gp.prior.kernel, X, params)
    L = jnp.linalg.cholesky(Kff + jnp.eye(N) * 1e-6)

    def build_rv(test_points: Array):
        Kfx = cross_covariance(gp.prior.kernel, X, test_points, params)
        Kxx = gram(gp.prior.kernel, test_points, params)
        A = solve_triangular(L, Kfx.T, lower=True)
        latent_var = Kxx - jnp.sum(jnp.square(A), -2)
        latent_mean = jnp.matmul(A.T, nu)
        lvar = jnp.diag(latent_var)
        moment_fn = predictive_moments(gp.likelihood)
        return moment_fn(latent_mean.ravel(), lvar)

    return build_rv


@dispatch(SpectralPosterior, dict, Dataset)
def random_variable(
    gp: SpectralPosterior,
    params: dict,
    training: Dataset,
    static_params: dict = None,
) -> tfd.Distribution:
    X, y = training.X, training.y

    params = concat_dictionaries(params, static_params)
    m = gp.prior.kernel.num_basis
    w = params["basis_fns"] / params["lengthscale"]
    phi = gp.prior.kernel._build_phi(X, params)

    A = (params["variance"] / m) * jnp.matmul(jnp.transpose(phi), phi) + params["obs_noise"] * I(
        2 * m
    )

    RT = jnp.linalg.cholesky(A)
    R = jnp.transpose(RT)

    RtiPhit = solve_triangular(RT, jnp.transpose(phi))
    Rtiphity = jnp.matmul(RtiPhit, y)

    alpha = params["variance"] / m * solve_triangular(R, Rtiphity, lower=False)

    def build_rv(test_points: Array):
        N = test_points.shape[0]
        phistar = jnp.matmul(test_points, jnp.transpose(w))
        phistar = jnp.hstack([jnp.cos(phistar), jnp.sin(phistar)])
        mean = jnp.matmul(phistar, alpha)

        RtiPhistart = solve_triangular(RT, jnp.transpose(phistar))
        PhiRistar = jnp.transpose(RtiPhistart)
        cov = (
            params["obs_noise"]
            * params["variance"]
            / m
            * jnp.matmul(PhiRistar, jnp.transpose(PhiRistar))
            + I(N) * 1e-6
        )
        return tfd.MultivariateNormalFullCovariance(mean.squeeze(), cov)

    return build_rv


@dispatch(jnp.DeviceArray, Prior, dict, Dataset)
def sample(
    key: jnp.DeviceArray, gp: Prior, params: dict, training: Dataset, n_samples: int = 1
) -> Array:
    rv = random_variable(gp, params, training)
    return rv.sample(sample_shape=(n_samples,), seed=key)


@dispatch(jnp.DeviceArray, Prior, dict, Array)
def sample(key: jnp.DeviceArray, gp: Prior, params: dict, X: Array, n_samples: int = 1) -> Array:
    training = Dataset(X=X, y=jnp.ones_like(X))
    rv = random_variable(gp, params, training)
    return rv.sample(sample_shape=(n_samples,), seed=key)


@dispatch(jnp.DeviceArray, tfd.Distribution)
def sample(key: jnp.DeviceArray, random_variable: tfd.Distribution, n_samples: int = 1) -> Array:
    return random_variable.sample(sample_shape=(n_samples,), seed=key)



from typing import Optional

from chex import dataclass
from multipledispatch import dispatch

from .kernels import Kernel
from .kernels.spectral import SpectralKernel
from .likelihoods import (Gaussian, Likelihood, NonConjugateLikelihoods,
                          NonConjugateLikelihoodType)
from .mean_functions import MeanFunction, Zero


@dataclass(repr=False)
class Prior:
    kernel: Kernel
    mean_function: Optional[MeanFunction] = Zero()
    name: Optional[str] = "Prior"

    @dispatch(Gaussian)
    def __mul__(self, other: Gaussian):
        if isinstance(self.kernel, SpectralKernel):
            return SpectralPosterior(prior=self, likelihood=other)
        else:
            return ConjugatePosterior(prior=self, likelihood=other)

    @dispatch(NonConjugateLikelihoods)
    def __mul__(self, other: NonConjugateLikelihoodType):
        return NonConjugatePosterior(prior=self, likelihood=other)


@dataclass
class Posterior:
    prior: Prior
    likelihood: Likelihood
    name: Optional[str] = "Posterior"


@dataclass
class ConjugatePosterior:
    prior: Prior
    likelihood: Gaussian
    name: Optional[str] = "ConjugatePosterior"

    def __repr__(self):
        meanf_string = self.prior.mean_function.__repr__()
        kernel_string = self.prior.kernel.__repr__()
        likelihood_string = self.likelihood.__repr__()
        return f"Conjugate Posterior\n{'-'*80}\n- {meanf_string}\n- {kernel_string}\n- {likelihood_string}"


@dataclass
class NonConjugatePosterior:
    prior: Prior
    likelihood: NonConjugateLikelihoodType
    name: Optional[str] = "ConjugatePosterior"

    def __repr__(self):
        meanf_string = self.prior.mean_function.__repr__()
        kernel_string = self.prior.kernel.__repr__()
        likelihood_string = self.likelihood.__repr__()
        return f"Conjugate Posterior\n{'-'*80}\n- {meanf_string}\n- {kernel_string}\n- {likelihood_string}"


@dataclass
class SpectralPosterior:
    prior: Prior
    likelihood: Gaussian
    name: Optional[str] = "SpectralPosterior"

    def __repr__(self):
        meanf_string = self.prior.mean_function.__repr__()
        kernel_string = self.prior.kernel.__repr__()
        likelihood_string = self.likelihood.__repr__()
        return f"Sparse Spectral Posterior\n{'-'*80}\n- {meanf_string}\n- {kernel_string}\n- {likelihood_string}"



from typing import Optional

import jax.numpy as jnp
from chex import dataclass
from jax import vmap
from multipledispatch import dispatch

from gpjax.types import Array

from .utils import scale, stretch


@dataclass(repr=False)
class Kernel:
    ndims: Optional[int] = 1
    stationary: Optional[bool] = False
    spectral: Optional[bool] = False
    name: Optional[str] = "Kernel"

    def __call__(self, x: Array, y: Array, params: dict) -> Array:
        raise NotImplementedError

    def __repr__(self):
        return f"{self.name}:\n\t Stationary: {self.stationary}\n\t Spectral form: {self.spectral} \n\t ARD structure: {self.ard}"

    @property
    def ard(self):
        return True if self.ndims > 1 else False


@dataclass(repr=False)
class RBF(Kernel):
    ndims: Optional[int] = 1
    stationary: Optional[bool] = True
    spectral: Optional[bool] = False
    name: Optional[str] = "Radial basis function kernel"

    def __call__(self, x: jnp.DeviceArray, y: jnp.DeviceArray, params: dict) -> Array:
        x = scale(x, params["lengthscale"])
        y = scale(y, params["lengthscale"])
        K = stretch(jnp.exp(-0.5 * squared_distance(x, y)), params["variance"])
        return K.squeeze()


@dispatch(RBF)
def initialise(kernel: RBF):
    return {"lengthscale": jnp.array([1.0] * kernel.ndims), "variance": jnp.array([1.0])}


def squared_distance(x: Array, y: Array):
    return jnp.sum((x - y) ** 2)


def gram(kernel: Kernel, inputs: Array, params: dict) -> Array:
    return vmap(lambda x1: vmap(lambda y1: kernel(x1, y1, params))(inputs))(inputs)


def cross_covariance(kernel: Kernel, x: Array, y: Array, params: dict) -> Array:
    return vmap(lambda x1: vmap(lambda y1: kernel(x1, y1, params))(x))(y)



from typing import Callable, Optional, Union

import jax.numpy as jnp
from chex import dataclass
from multipledispatch import dispatch
from tensorflow_probability.substrates.jax import distributions as tfd


@dataclass(repr=False)
class Likelihood:
    name: Optional[str] = "Likelihood"

    def __repr__(self):
        return f"{self.name} likelihood function"


@dataclass(repr=False)
class Gaussian(Likelihood):
    name: Optional[str] = "Gaussian"


@dispatch(Gaussian)
def initialise(likelihood: Gaussian) -> dict:
    return {"obs_noise": jnp.array([1.0])}


@dataclass(repr=False)
class Bernoulli(Likelihood):
    name: Optional[str] = "Bernoulli"


@dispatch(Bernoulli)
def initialise(likelihood: Bernoulli) -> dict:
    return {}


@dispatch(Bernoulli)
def link_function(likelihood: Bernoulli):
    return tfd.ProbitBernoulli


@dispatch(Bernoulli)
def predictive_moments(likelihood: Bernoulli) -> Callable:
    link = link_function(likelihood)

    def moments(mean: jnp.DeviceArray, variance: jnp.DeviceArray) -> tfd.Distribution:
        rv = link(mean / jnp.sqrt(1 + variance))
        return rv

    return moments


@dataclass(repr=False)
class Poisson(Likelihood):
    name: Optional[str] = "Poisson"


@dispatch(Poisson)
def initialise(likelihood: Poisson) -> dict:
    return {}


NonConjugateLikelihoods = (Bernoulli, Poisson)
NonConjugateLikelihoodType = Union[Bernoulli, Poisson]



from typing import Callable

import jax.numpy as jnp
import jax.random as jr
import pytest
from tensorflow_probability.substrates.jax import distributions as tfd

from gpjax import Dataset, Prior
from gpjax.config import get_defaults
from gpjax.kernels import RBF, to_spectral
from gpjax.likelihoods import Bernoulli, Gaussian
from gpjax.objectives import marginal_ll
from gpjax.parameters import build_all_transforms, initialise


def test_spectral():
    key = jr.PRNGKey(123)
    kern = to_spectral(RBF(), 10)
    posterior = Prior(kernel=kern) * Gaussian()
    x = jnp.linspace(-1.0, 1.0, 20).reshape(-1, 1)
    y = jnp.sin(x)
    D = Dataset(X=x, y=y)
    params = initialise(key, posterior)
    config = get_defaults()
    unconstrainer, constrainer = build_all_transforms(params.keys(), config)
    params = unconstrainer(params)
    mll = marginal_ll(posterior, transform=constrainer)
    assert isinstance(mll, Callable)
    neg_mll = marginal_ll(posterior, transform=constrainer, negative=True)
    assert neg_mll(params, D) == jnp.array(-1.0) * mll(params, D)
    nmll = neg_mll(params, D)
    assert nmll.shape == ()



from typing import Callable

import jax.numpy as jnp
from jax.scipy.linalg import cho_factor, cho_solve, cholesky, solve_triangular
from multipledispatch import dispatch

from .gps import ConjugatePosterior, NonConjugatePosterior, Prior
from .kernels import cross_covariance, gram
from .likelihoods import predictive_moments
from .types import Array, Dataset
from .utils import I


@dispatch(ConjugatePosterior, dict, Dataset)
def mean(gp: ConjugatePosterior, param: dict, training: Dataset) -> Callable:
    X, y = training.X, training.y
    sigma = param["obs_noise"]
    n_train = training.n
    # Precompute covariance matrices
    Kff = gram(gp.prior.kernel, X, param)
    prior_mean = gp.prior.mean_function(X)
    L = cho_factor(Kff + I(n_train) * sigma, lower=True)

    prior_distance = y - prior_mean
    weights = cho_solve(L, prior_distance)

    def meanf(test_inputs: Array) -> Array:
        Kfx = cross_covariance(gp.prior.kernel, X, test_inputs, param)
        return jnp.dot(Kfx, weights)

    return meanf


@dispatch(ConjugatePosterior, dict, Dataset)
def variance(
    gp: ConjugatePosterior,
    param: dict,
    training: Dataset,
) -> Callable:
    X, y = training.X, training.y
    sigma = param["obs_noise"]
    n_train = training.n
    Kff = gram(gp.prior.kernel, X, param)
    L = cho_factor(Kff + I(n_train) * sigma, lower=True)

    def varf(test_inputs: Array) -> Array:
        Kfx = cross_covariance(gp.prior.kernel, X, test_inputs, param)
        Kxx = gram(gp.prior.kernel, test_inputs, param)
        latents = cho_solve(L, Kfx.T)
        return Kxx - jnp.dot(Kfx, latents)

    return varf


@dispatch(NonConjugatePosterior, dict, Dataset)
def mean(gp: NonConjugatePosterior, param: dict, training: Dataset) -> Array:
    ell, alpha, nu = param["lengthscale"], param["variance"], param["latent"]
    X, y = training.X, training.y
    N = training.n
    Kff = gram(gp.prior.kernel, X, param)
    L = jnp.linalg.cholesky(Kff + I(N) * 1e-6)

    def meanf(test_inputs: Array) -> Array:
        Kfx = cross_covariance(gp.prior.kernel, X, test_inputs, param)
        Kxx = gram(gp.prior.kernel, test_inputs, param)
        A = solve_triangular(L, Kfx.T, lower=True)
        latent_var = Kxx - jnp.sum(jnp.square(A), -2)
        latent_mean = jnp.matmul(A.T, nu)

        lvar = jnp.diag(latent_var)

        moment_fn = predictive_moments(gp.likelihood)
        pred_rv = moment_fn(latent_mean.ravel(), lvar)
        return pred_rv.mean()

    return meanf


@dispatch(NonConjugatePosterior, dict, Dataset)
def variance(gp: NonConjugatePosterior, param: dict, training: Dataset):
    X, y = training.X, training.y
    N = training.n
    ell, alpha, nu = param["lengthscale"], param["variance"], param["latent"]
    Kff = gram(gp.prior.kernel, X, param)
    L = jnp.linalg.cholesky(Kff + I(N) * 1e-6)

    def variancef(test_inputs: Array) -> Array:
        Kfx = cross_covariance(gp.prior.kernel, X, test_inputs, param)
        Kxx = gram(gp.prior.kernel, test_inputs, param)
        A = solve_triangular(L, Kfx.T, lower=True)
        latent_var = Kxx - jnp.sum(jnp.square(A), -2)
        latent_mean = jnp.matmul(A.T, nu)
        lvar = jnp.diag(latent_var)
        moment_fn = predictive_moments(gp.likelihood)
        pred_rv = moment_fn(latent_mean.ravel(), lvar)
        return pred_rv.variance()

    return variancef



from typing import Optional

import jax.numpy as jnp
from chex import dataclass
from multipledispatch import dispatch

from .types import Array


@dataclass(repr=False)
class MeanFunction:
    output_dim: Optional[int] = 1
    name: Optional[str] = "Mean function"

    def __call__(self, x: Array) -> Array:
        raise NotImplementedError

    def __repr__(self):
        return f"{self.name}\n\t Output dimension: {self.output_dim}"


@dataclass(repr=False)
class Zero(MeanFunction):
    output_dim: Optional[int] = 1
    name: Optional[str] = "Zero mean function"

    def __call__(self, x: Array) -> Array:
        out_shape = (x.shape[0], self.output_dim)
        return jnp.zeros(shape=out_shape)


@dispatch(Zero)
def initialise(meanf: Zero) -> dict:
    return {}



import tensorflow_probability.substrates.jax.bijectors as tfb
from ml_collections import ConfigDict


def get_defaults():
    config = ConfigDict()
    # Covariance matrix stabilising jitter
    config.jitter = 1e-6
    # Parameter transformations
    config.transformations = transformations = ConfigDict()
    transformations.positive_transform = tfb.Softplus
    transformations.identity_transform = tfb.Identity

    transformations.lengthscale = "positive_transform"
    transformations.variance = "positive_transform"
    transformations.obs_noise = "positive_transform"
    transformations.latent = "identity_transform"
    transformations.basis_fns = "identity_transform"
    return config
